module TypeHelperFunctions exposing (..)

import Grid exposing (Grid)
import List.Extra
import List exposing (map)
import Types exposing (..)
import Tetros exposing (..)
import GeneralHelperFunctions exposing (..)

createEmptyShape : ShapeStatus
createEmptyShape = {x=0,y=0,v=0,sIdent=-1,placed=False,sType=None}

getLength : LevelList -> Int
getLength list = case list of
    Levels x -> List.length x
    Editor (lvlInfo, estate) -> List.length lvlInfo.levels
    _ -> 0

extractLevelList : LevelList -> List (List ShapeStatus, Grid Bool)
extractLevelList list = case list of
    Levels l -> l
    _ -> []

getXY : ShapeStatus -> (Int, Int)
getXY shape = (shape.x, shape.y)

getShorthand : Maybe (List CustomShape) -> String -> Maybe CustomShape
getShorthand cs x = case cs of
    Nothing -> Nothing
    Just jcs -> case List.head jcs of
        Nothing -> Nothing
        Just e -> if e.shortHand == x then Just e else (getShorthand (List.tail jcs) x)

getShortHandA : AbstractShape -> String
getShortHandA ashape = case ashape of
    I -> "I"
    O -> "O"
    J -> "J"
    L -> "L"
    T -> "T"
    S -> "S"
    Z -> "Z"
    None -> "None"
    Custom shape -> shape.shortHand

getShorthandS : CustomShape -> String
getShorthandS shape = shape.shortHand

removeFromBoard : ShapeStatus -> ShapeStatus
removeFromBoard s = {s | placed = False}

placeOnBoard : ShapeStatus -> ShapeStatus
placeOnBoard s = {s | placed = True}

getShapeColour : AbstractShape -> String
getShapeColour shape = case shape of
    L -> "orange"
    J -> "blue"
    S -> "green"
    Z -> "red"
    I -> "cyan"
    T -> "purple"
    O -> "yellow"
    None -> "black"
    Custom _ -> "black"

getShapePoints : AbstractShape -> List (Int, Int)
getShapePoints shape = case shape of
    L -> lShape
    J -> jShape
    S -> sShape
    Z -> zShape
    I -> iShape
    T -> tShape
    O -> oShape
    None -> []
    Custom _ -> []

indexLL : LevelList -> Int -> (List (ShapeStatus), Grid Bool)
indexLL levels num = case levels of
    Levels list -> case List.Extra.getAt num list of
        Just x -> x
        Nothing -> ([], Grid.repeat 1 1 False)
    _ -> ([], Grid.repeat 1 1 False)

getRealShapes : (List (AbstractShape), Level) -> (List (ShapeStatus), Level)
getRealShapes (shapes, lvl) = (List.map2 createShapeStatus shapes (List.range 1 (List.length shapes)), lvl)

createShapeStatus : AbstractShape -> Int -> ShapeStatus
createShapeStatus shape num = {x=(generateShapeX shape num), y=(generateShapeY shape num), v=0 ,sIdent=num, placed=False, sType=shape}

generateShapeX : AbstractShape -> Int -> Int
generateShapeX s num = if num > 8
    then (730+getShapeXTranform s) + if modBy 2 num == 0 then 0 else 80
    else (30+getShapeXTranform s) + if modBy 2 num == 0 then 0 else 80

generateShapeY : AbstractShape -> Int -> Int
generateShapeY s num = let modNum = modBy 8 num in 60+((500//9)*modNum)

getShapeXTranform : AbstractShape -> Int
getShapeXTranform s = case s of
    Z -> 150
    J -> 150
    _ -> 0

adjustToCenter : (Int, Int) -> (Float, Float)
adjustToCenter (x,y) = (toFloat x+0.5, toFloat y+0.5)

adjustToOrigin : (Float, Float) -> (Int, Int)
adjustToOrigin (fx,fy) = (round (fx-0.5), round (fy-0.5))

rotateTupleCw : (Float, Float) -> (Float, Float)
rotateTupleCw (x,y) = (y*(-1),x*(1))

applyNumbers : List (a, Int) -> List (a, Int)
applyNumbers ls = List.map2 (\newI (a,oldI) -> (a,newI)) (List.range 1 (List.length ls)) ls

makeCShapeIndex : List (CustomShape, Int) -> (CustomShape, Maybe Int) -> List (CustomShape, Int)
makeCShapeIndex customShapes (cshape, mindex) = case mindex of
    Nothing ->
        let
            newCShape = {cshape | polygonPoints = pNormalize cshape.polygonPoints, coordinates = pNormalize cshape.coordinates}
        in customShapes++[(newCShape, List.length customShapes +1)]
    Just index -> List.Extra.setAt (index-1) (cshape, index) customShapes

toAbstractShapes : Maybe (List CustomShape) -> Level -> (List (AbstractShape), Level)
toAbstractShapes mcs levelObj =
    let
        customs = case mcs of
            Nothing -> []
            Just cs -> cs
    in (List.concatMap (stringToShape customs) levelObj.solution, levelObj)

levelToGrid : (List (ShapeStatus), Level) -> (List ShapeStatus, Grid Bool)
levelToGrid (list, lvl) =
    let
        bx = case lvl.boardx of
            Just x -> x
            Nothing -> 8
        by = case lvl.boardy of
            Just y -> y
            Nothing -> 8
    in (list, Grid.repeat bx by False)

stringToShape : List CustomShape -> String -> List AbstractShape
stringToShape cs str = case str of
    "I" -> [I]
    "T" -> [T]
    "O" -> [O]
    "J" -> [J]
    "L" -> [L]
    "S" -> [S]
    "Z" -> [Z]
    x -> case getShorthand (Just cs) x of
        Nothing -> []
        Just elem -> [Custom elem]

getShapeFromShortHand : List AbstractShape -> (Int, String) -> (AbstractShape,Int)
getShapeFromShortHand shapes (i,s) = case s of
    "L" -> (L, i)
    "J" -> (J, i)
    "I" -> (I, i)
    "O" -> (O, i)
    "T" -> (T, i)
    "S" -> (S, i)
    "Z" -> (Z, i)
    _ -> case List.head (List.filter (((==) (s)) << getShorthandS) (map abstractToCustom shapes)) of
            Just shape -> (Custom shape, i)
            Nothing -> (Custom {canRotate = False, polygonPoints = [], coordinates = [], colour= "#000000", shortHand = "err"},0)

getLevelSolution : Int -> List EditLevel -> List String
getLevelSolution i levels = case List.Extra.getAt (i-1) levels of
    Just lvl -> lvl.solution
    Nothing -> []

getAbstractCoordinates : AbstractShape -> List (Int, Int)
getAbstractCoordinates ashape = case ashape of
    Custom shape -> shape.coordinates
    None -> []
    x -> [(0,0),(0,0),(0,0),(0,0)]

abstractToCustom : AbstractShape -> CustomShape
abstractToCustom ashape = case ashape of
    Custom shape -> shape
    _ -> {canRotate = False, polygonPoints = [], coordinates = [], colour= "#000000", shortHand = "err"}

customToAbstract : (CustomShape, Int) -> (AbstractShape, Int)
customToAbstract (shape, i) = (Custom shape, i)

getHeight : Model -> Int
getHeight model = case model.windowHeight of
    Nothing -> 0
    Just h -> h

pNormalize : List (Int, Int) -> List (Int, Int)
pNormalize polygonPoints =
    let
        (x, y) = (List.minimum (map Tuple.first polygonPoints), List.minimum (map Tuple.second polygonPoints))
    in case (x, y) of
        (Just xmin, Just ymin) -> map (flip subTuple (xmin,ymin)) polygonPoints
        _ -> polygonPoints

getRotatable : Model -> Bool
getRotatable model = case model.levelList of
    Editor (_, ShapeCreator (shape,_)) -> shape.canRotate
    _ -> False

getColour : Model -> String
getColour model = case model.levelList of
    Editor (_, ShapeCreator (shape,_)) -> shape.colour
    _ -> ""

getShapeName : Model -> String
getShapeName model = case model.levelList of
    Editor (_, ShapeCreator (shape,_)) -> shape.shortHand
    _ -> ""

getWidthE : Model -> String
getWidthE model = case model.levelList of
    Editor (lvlInfo, levelNum) -> case List.Extra.getAt (model.currentLevel-1) lvlInfo.levels of
        Just lvl -> String.fromInt lvl.boardx
        Nothing -> ""
    _ -> ""

getHeightE : Model -> String
getHeightE model = case model.levelList of
    Editor (lvlInfo, levelNum) -> case List.Extra.getAt (model.currentLevel-1) lvlInfo.levels of
        Just lvl -> String.fromInt lvl.boardy
        Nothing -> ""
    _ -> ""
