module Types exposing (..)

import File exposing (File)
import Grid exposing (Grid)
import Browser.Navigation as Nav
import Url
import Browser
import Http

type alias Model =
  { mpos : (Int, Int)
  , key : Nav.Key
  , url : Url.Url
  , shapeStatus : ShapeStatus
  , saved : List (ShapeStatus)
  , boardGrid : Grid Bool
  , boardIsFull : Bool
  , mousePiecePosition : (Int, Int)
  , levelList : LevelList
  , currentLevel : Int
  , completedLevels : List (Int)
  , customLevels : String
  , jsonHelpState : Maybe Bool
  , exampleLevels : String
  , exampleResult : Maybe Bool
  , darkMode : Bool
  , exitModalState : Bool
  , drawState : Maybe DrawState
  , windowHeight : Maybe Int}

type DrawState = DPolygon | DCoordinates

type alias Level = { 
      boardx : Maybe Int
    , boardy : Maybe Int
    , solution : List String}

type alias EditLevel = {
      boardx: Int
    , boardy: Int
    , solution: List String
    }

type alias EditLevelInfo =
    { customShapes : List (CustomShape, Int)
    , levels : List EditLevel}

type alias LevelInfo =
    { customShapes : Maybe (List CustomShape)
    , levels : List Level}

type LevelList = Empty | Loading | Failed | Levels (List ((List ShapeStatus), Grid Bool)) | Editor (EditLevelInfo, EditState)

type EditState = ShapeCreator (CustomShape, Maybe Int) | LevelNumber

type AbstractShape = L | J | T | O | I | S | Z | Custom CustomShape | None

type alias CustomShape =
    { canRotate : Bool
    , shortHand : String
    -- Wie ist die Kollision?
    , coordinates : List (Int, Int)
    -- Wie wird das Polyomino gezeichnet?
    , polygonPoints : List (Int, Int)
    , colour : String
    }

type alias ShapeStatus =
    { x : Int
    , y : Int
    , v : Int
    , sIdent : Int
    , placed : Bool
    , sType : AbstractShape
    }

type Msg =
    Position Int Int
  | LinkClicked Browser.UrlRequest
  | UrlChanged Url.Url
  | RightRotate
  | PlaceShape
  | TakeShape Int
  | GotLevels (Result Http.Error (LevelInfo))
  | LoadLevels
  | LoadCustomLevels
  | CustomLevelsSelected File
  | CustomLevelsLoaded String
  | DecodeCustomLevels
  | CompleteLevel
  | SwitchJsonHelp
  | ExampleChanged String
  | LoadExample
  | ToggleTheme
  | DisableModal
  | EnableModal
  | ToMainMenu
  | EnterEditor
  | CheckRotate Bool
  | NewColour String
  | ChangeName String
  | DrawCoordinates
  | DrawPolygon
  | ResetPolygon
  | AddCoordinate (Int, Int)
  | RemoveCoordinate (Int, Int)
  | AddPolygonPoint (Int, Int)
  | UndoPolygonPoint
  | SaveCustomShape
  | NewPiece
  | LoadShape Int
  | RemoveShape Int
  | RemoveTetroShape AbstractShape
  | ToEditor
  | ChangeSizeX String
  | ChangeSizeY String
  | AddShape Int
  | AddTetroShape AbstractShape
  | DeleteLevel
  | SaveLevels
  | ChangeHeight Int

type alias MousePosition = { x : Float, y : Float }
